import warnings, types, pickle, marshal, numpy as np
from tqdm.auto import tqdm
from joblib import Parallel, delayed
from multiprocessing import cpu_count
from dtaidistance import dtw, dtw_ndim
from sklearn.metrics import confusion_matrix
from sklearn.preprocessing import LabelEncoder
from ...internals import _Validator

dtw_cc = None
try:
    from dtaidistance import dtw_cc
except ImportError:
    pass

class KNNClassifier:
    """A k-Nearest Neighbor classifier that uses dynamic time warping as a distance measure for comparing observation sequences of different length.

    Parameters
    ----------
    k: int > 0
        Number of neighbors.

    classes: array-like of str/numeric
        The complete set of possible classes/labels.

    weighting: callable, optional
        A callable that specifies how distance weighting should be performed.
        The callable should accept a :class:`numpy:numpy.ndarray` of DTW distances, apply an element-wise weighting transformation,
        then return an equally-sized :class:`numpy:numpy.ndarray` of weighted distances.

        If no weighting is chosen then the function ``lambda x: np.ones_like(x)`` is used, which weights all of the distances equally.

        Examples of weighting functions are:

        - :math:`e^{-\\alpha x}`, specified by ``lambda x: np.exp(-alpha * x)`` for some positive :math:`\\alpha`,
        - :math:`\\frac{1}{x}`, specified by ``lambda x: 1 / x``.

        A good weighting function should *ideally* be defined at :math:`x=0` in the event that two observations are perfectly aligned and therefore have zero DTW distance.

        .. tip::
            It may be desirable to restrict DTW distances to a small range if you intend to use a weighting function.

            Using the :class:`~MinMaxScale` or :class:`~Standardize` preprocessing transformations to scale your features helps to ensure that DTW distances remain small.

    window: 0 ≤ float ≤ 1
        The width of the Sakoe-Chiba band global constraint as a fraction of the length of the longest of the two sequences.
        A larger constraint will speed up the DTW alignment by restricting the maximum temporal deviation from the diagonal of the DTW matrix,
        but too much constraint may lead to poor alignment.

        The default value of 1 corresponds to full DTW computation with no global constraint applied.

    use_c: bool
        Whether or not to use fast pure C compiled functions (from the `dtaidistance <https://github.com/wannesm/dtaidistance>`_ package) to perform the DTW computations.

        .. tip::
            If you set ``use_c = True`` and are receiving an error about a C library not being available, try reinstalling ``dtaidistance`` and disabling the cache:

            .. code-block:: console

                pip install -vvv --upgrade --no-cache-dir --force-reinstall dtaidistance

    independent: bool
        Whether or not to allow features to be warped independently from each other. See `here <https://www.cs.ucr.edu/~eamonn/Multi-Dimensional_DTW_Journal.pdf>`_ for a good overview of both approaches.

    random_state: numpy.random.RandomState, int, optional
        A random state object or seed for reproducible randomness.

    Attributes
    ----------
    k (property): int > 0
        The number of neighbors.

    weighting (property): callable
        The distance weighting function.

    window (property): 0 ≤ float ≤ 1
        The width of the Sakoe-Chiba band global constraint as a fraction of the length of the longest of the two sequences.

    use_c (property): bool
        Whether or not to use fast pure C compiled functions to perform the DTW computations.

    encoder_ (property): sklearn.preprocessing.LabelEncoder
        The label encoder fitted on the set of ``classes`` provided during instantiation.

    classes_ (property): numpy.ndarray (str/numeric)
        The complete set of possible classes/labels.

    X_ (property): list of numpy.ndarray (float)
        A list of multiple observation sequences used to fit the classifier.

    y_ (property): numpy.ndarray (int)
        The encoded labels for the observation sequences used to fit the classifier.
    """

    def __init__(self, k, classes, weighting=None, window=1., use_c=False, independent=False, random_state=None):
        self._val = _Validator()
        self._k = self._val.is_restricted_integer(
            k, lambda x: x > 0, desc='number of neighbors', expected='greater than zero')
        self._window = float(window) if window in (0, 1) else self._val.is_restricted_float(
            window, lambda x: 0. <= x <= 1., desc='Sakoe-Chiba band width (fraction)', expected='between zero and one')
        self._random_state = self._val.is_random_state(random_state)

        self._val.is_iterable(classes, 'classes')
        self._val.is_string_or_numeric(classes[0], 'each class')
        if all(isinstance(label, type(classes[0])) for label in classes[1:]):
            self._encoder_ = LabelEncoder().fit(classes)
        else:
            raise TypeError('Expected all classes to be of the same type')

        self._weighting = self._val.is_func(weighting, 'distance weighting function') if weighting else lambda x: np.ones_like(x)

        self._use_c = self._val.is_boolean(use_c, desc='whether or not to use fast pure C compiled functions')
        if self._use_c and (dtw_cc is None):
            warnings.warn('DTAIDistance C library not available – using Python implementation', ImportWarning)
            self._use_c = False

        self._independent = self._val.is_boolean(independent, 'independent')
        self._dtw = self._dtwi if independent else self._dtwd

    def fit(self, X, y):
        """Fits the classifier by adding labeled training observation sequences.

        Parameters
        ----------
        X: list of numpy.ndarray (float)
            A list of multiple observation sequences.

        y: array-like of str/numeric
            An iterable of labels for the observation sequences.
        """
        X, y = self._val.is_observation_sequences_and_labels(X, y, dtype=np.float64)
        self._X_, self._y_ = X, self._encoder_.transform(y)
        self._n_features_ = X[0].shape[1]
        return self

    def predict(self, X, return_scores=False, original_labels=True, verbose=True, n_jobs=1):
        """Predicts the label for an observation sequence (or multiple sequences).

        Parameters
        ----------
        X: numpy.ndarray (float) or list of numpy.ndarray (float)
            An individual observation sequence or a list of multiple observation sequences.

        return_scores: bool
            Whether to return the scores for each class on the observation sequence(s).

        original_labels: bool
            Whether to inverse-transform the labels to their original encoding.

        verbose: bool
            Whether to display a progress bar or not.

            .. note::
                The progress bar cannot be displayed if both ``verbose=True`` and ``n_jobs > 1``.

        n_jobs: int > 0 or -1
            | The number of jobs to run in parallel.
            | Setting this to -1 will use all available CPU cores.

        Returns
        -------
        prediction(s): str/numeric or :class:`numpy:numpy.ndarray` (str/numeric)
            The predicted label(s) for the observation sequence(s).

            If ``original_labels`` is true, then the returned labels are
            inverse-transformed into their original encoding.
        """
        (self.X_, self.y_)
        X = self._val.is_observation_sequences(X, allow_single=True, dtype=np.float64)
        self._val.is_boolean(original_labels, desc='original_labels')
        self._val.is_boolean(verbose, desc='verbose')
        self._val.is_restricted_integer(n_jobs, lambda x: x == -1 or x > 0, 'number of jobs', '-1 or greater than zero')

        n_jobs = min(cpu_count() if n_jobs == -1 else n_jobs, len(X))

        # Make prediction for a single sequence
        if isinstance(X, np.ndarray):
            if n_jobs > 1:
                warnings.warn('Single predictions do not yet support multi-processing. Set n_jobs=1 to silence this warning.')

            output = self._predict(X, verbose=verbose)

        # Make predictions for multiple sequences (in parallel)
        else:
            if n_jobs == 1:
                # Use entire list as a chunk
                output = self._chunk_predict(X, verbose=verbose)

            else:
                if verbose:
                    warnings.warn('Progress bars cannot be displayed when using multiple processes. Set verbose=False to silence this warning.')

                # Split X into n_jobs equally sized chunks and process in parallel
                chunks = [list(chunk) for chunk in np.array_split(np.array(X, dtype=object), n_jobs)]
                output = np.vstack(Parallel(n_jobs=n_jobs)(delayed(self._chunk_predict)(chunk) for chunk in chunks))

        return self._output(output, return_scores=return_scores, original_labels=original_labels)

    def evaluate(self, X, y, verbose=True, n_jobs=1):
        """Evaluates the performance of the classifier on a batch of observation sequences and their labels.

        Parameters
        ----------
        X: list of numpy.ndarray (float)
            A list of multiple observation sequences.

        y: array-like of str/numeric
            An iterable of labels for the observation sequences.

        verbose: bool
            Whether to display a progress bar for predictions or not.

        n_jobs: int > 0 or -1
            | The number of jobs to run in parallel.
            | Setting this to -1 will use all available CPU cores.

        Returns
        -------
        accuracy: float
            The categorical accuracy of the classifier on the observation sequences.

        confusion: :class:`numpy:numpy.ndarray` (int)
            The confusion matrix representing the discrepancy between predicted and actual labels.
        """
        X, y = self._val.is_observation_sequences_and_labels(X, y, dtype=np.float64)
        self._val.is_boolean(verbose, desc='verbose')
        predictions = self.predict(X, return_scores=False, original_labels=False, verbose=verbose, n_jobs=n_jobs)
        cm = confusion_matrix(self._encoder_.transform(y), predictions, labels=self._encoder_.transform(self._encoder_.classes_))
        return np.sum(np.diag(cm)) / np.sum(cm), cm

    def save(self, path):
        """Serializes the :class:`KNNClassifier` object by pickling its hyper-parameters, variables and training data.

            .. warning::
                As :math:`k`-NN classifier must look through each training example during prediction,
                saving the classifier simply copies all of the training observation sequences and labels.

        Parameters
        ----------
        path: str
            File path (usually with `.pkl` extension) to store the serialized :class:`KNNClassifier` object.
        """
        (self.X_, self.y_)
        with open(path, 'wb') as file:
            pickle.dump({
                'k': self._k,
                'classes': self._encoder_.classes_,
                'weighting': marshal.dumps((self._weighting.__code__, self._weighting.__name__)),
                'window': self._window,
                'use_c': self._use_c,
                'independent': self._independent,
                'random_state': self._random_state,
                'X': self._X_,
                'y': self._y_,
                'n_features': self._n_features_
            }, file)

    @classmethod
    def load(cls, path):
        """Deserializes a :class:`KNNClassifier` object which was serialized with the :meth:`save` function.

        Parameters
        ----------
        path: str
            File path of the serialized data generated by the :meth:`save` method.

        Returns
        -------
        deserialized: :class:`KNNClassifier`
            The deserialized DTW :math:`k`-NN classifier object.
        """
        with open(path, 'rb') as file:
            data = pickle.load(file)

        # Check deserialized object dictionary and keys
        keys = set(('k', 'classes', 'weighting', 'window', 'use_c', 'independent', 'random_state', 'X', 'y', 'n_features'))
        if not isinstance(data, dict):
            raise TypeError('Expected deserialized object to be a dictionary - make sure the object was serialized with the save() function')
        else:
            if len(set(keys) - set(data.keys())) != 0:
                raise ValueError('Missing keys in deserialized object dictionary – make sure the object was serialized with the save() function')

        # Deserialize the weighting function
        weighting, name = marshal.loads(data['weighting'])
        weighting = types.FunctionType(weighting, globals(), name)

        # Instantiate a new KNNClassifier with the same hyper-parameters
        clf = cls(
            k=data['k'],
            classes=data['classes'],
            weighting=weighting,
            window=data['window'],
            use_c=data['use_c'],
            independent=data['independent'],
            random_state=data['random_state']
        )

        # Load the data directly
        clf._X_, clf._y_ = data['X'], data['y']
        clf._n_features_ = data['n_features']

        return clf

    def _dtw_1d(self, a, b, window):
        """Computes the DTW distance between two univariate sequences."""
        return dtw.distance(a, b, use_c=self._use_c, window=window)

    def _dtwi(self, A, B):
        """Computes the multivariate DTW distance as the sum of the pairwise per-feature DTW distances, allowing each feature to be warped independently."""
        window = max(1, int(self._window * max(len(A), len(B))))
        return np.sum([self._dtw_1d(A[:, i], B[:, i], window=window) for i in range(self._n_features_)])

    def _dtwd(self, A, B):
        """Computes the multivariate DTW distance so that the warping of the features depends on each other, by modifying the local distance measure."""
        window = max(1, int(self._window * max(len(A), len(B))))
        return dtw_ndim.distance(A, B, use_c=self._use_c, window=window)

    def _multi_argmax(self, arr):
        """Same as numpy.argmax but returns all occurrences of the maximum and only requires a single pass.
        From: https://stackoverflow.com/a/58652335
        """
        all_, max_ = [0], arr[0]
        for i in range(1, len(arr)):
            if arr[i] > max_:
                all_, max_ = [i], arr[i]
            elif arr[i] == max_:
                all_.append(i)
        return np.array(all_)

    def _find_k_nearest(self, distances):
        """Returns the labels and weightings (or scores) of the k-nearest neighbors"""
        idx = np.argpartition(distances, self._k)[:self._k]
        return self._y_[idx], self._weighting(distances[idx])

    def _find_max_labels(self, nearest_labels, nearest_scores):
        """Returns the mode label(s) of the k nearest neighbors.
        Vectorization from: https://stackoverflow.com/a/49239335
        """
        # Sort the labels in ascending order (and sort distances in the same order)
        sorted_labels_idx = nearest_labels.argsort()
        sorted_labels, sorted_scores = nearest_labels[sorted_labels_idx], nearest_scores[sorted_labels_idx]
        # Identify the indices where the sorted labels change (so we can group by labels)
        change_idx = np.concatenate(([0], np.nonzero(np.diff(sorted_labels))[0] + 1))
        # Calculate the total score for each label
        label_scores = np.add.reduceat(sorted_scores, change_idx)
        # Find the change index of the maximum score(s)
        max_score_idx = change_idx[self._multi_argmax(label_scores)]
        # Map the change index of the maximum scores back to the actual label(s)
        max_labels = sorted_labels[max_score_idx]
        # Store class scores
        scores = np.full(len(self.classes_), -np.inf)
        scores[sorted_labels[change_idx]] = label_scores
        # Map the change index of the maximum scores back to the actual label(s), and return scores
        return max_labels, scores

    def _predict(self, x1, verbose=False):
        """Makes a prediction for a single observation sequence."""
        # Calculate DTW distances between x1 and all other sequences
        distances = np.array([self._dtw(x1, x2) for x2 in tqdm(self._X_, desc='Calculating distances', disable=not(verbose))])
        # Find the k-nearest neighbors by DTW distance
        nearest_labels, nearest_scores = self._find_k_nearest(distances)
        # Out of the k-nearest neighbors, find the label(s) which had the highest total weighting
        max_labels, scores = self._find_max_labels(nearest_labels, nearest_scores)
        # Randomly pick from the set of labels with the maximum label score
        label = self._random_state.choice(max_labels, size=1)
        # Combine the label with the scores
        return np.concatenate((label, scores))

    def _chunk_predict(self, chunk, verbose=False):
        """Makes predictions for multiple observation sequences."""
        return np.array([self._predict(x, verbose=False) for x in tqdm(chunk, desc='Predicting', disable=not(verbose))])

    def _output(self, output, return_scores, original_labels):
        """Splits the label from the scores, inverse-transforms the labels if necessary, and returns the result."""
        if output.ndim == 1:
            labels, scores = int(output[0]), output[1:]
            labels = self._encoder_.inverse_transform([labels]).item() if original_labels else labels
        else:
            labels, scores = output[:, 0].astype(int), output[:, 1:]
            labels = self._encoder_.inverse_transform(labels) if original_labels else labels
        return (labels, scores) if return_scores else labels

    @property
    def k(self):
        return self._k

    @property
    def weighting(self):
        return self._weighting

    @property
    def window(self):
        return self._window

    @property
    def use_c(self):
        return self._use_c

    @property
    def encoder_(self):
        return self._encoder_

    @property
    def classes_(self):
        return self._encoder_.classes_

    @property
    def X_(self):
        try:
            return self._X_
        except AttributeError:
            raise RuntimeError('The classifier needs to be fitted first')

    @property
    def y_(self):
        try:
            return self._y_
        except AttributeError:
            raise RuntimeError('The classifier needs to be fitted first')

    def __repr__(self):
        name = '.'.join([self.__class__.__module__.split('.')[0], self.__class__.__name__])
        attrs = [
            ('k', repr(self._k)),
            ('window', repr(self._window)),
            ('use_c', repr(self._use_c)),
            ('independent', repr(self._independent)),
            ('classes', repr(list(self._encoder_.classes_)))
        ]
        return '{}({})'.format(name, ', '.join('{}={}'.format(name, val) for name, val in attrs))


from sequentia.datasets import load_digits

def test_load_digits_full():
    numbers = range(10)
    dataset = load_digits(numbers=numbers)
    assert len(dataset.X) == 3000
    assert len(dataset.y) == 3000
    assert all(((x >= -1) & (x <= 1)).all() for x in dataset.X)
    assert set(dataset.y) == set(numbers)
    assert all((dataset.y == c).sum() == 300 for c in numbers)

def test_load_digits_subset():
    numbers = (0, 2, 6, 7)
    dataset = load_digits(numbers=numbers)
    assert len(dataset.X) == 1200
    assert len(dataset.y) == 1200
    assert all(((x >= -1) & (x <= 1)).all() for x in dataset.X)
    assert set(dataset.y) == set(numbers)
    assert all((dataset.y == c).sum() == 300 for c in numbers)


import pytest, numpy as np
from itertools import product
from sequentia.datasets.random_sequences import load_random_sequences, _sample_prior, _K, _sample_from_range
from ...support import assert_equal

random_state = np.random.RandomState(0)

K = lambda x1, x2: _K(x1, x2, variance=1, lengthscale=1)

# load_random_sequences

@pytest.mark.parametrize(
    'n_sequences,n_features,n_classes,length_range',
    list(product(
        [1, 10, 25, 30],
        [1, 2, 3],
        [1, 2, 5],
        [(5, 15), 15]
    ))
)
def test_load_random_sequences(n_sequences, n_features, n_classes, length_range):
    if n_classes > n_sequences:

        with pytest.raises(Exception) as e:
            load_random_sequences(
                n_sequences, n_features, n_classes, length_range,
                random_state=random_state
            )

        assert str(e.value) == 'Expected number of classes to be no more than n_sequences'

    else:

        dataset = load_random_sequences(
            n_sequences, n_features, n_classes, length_range,
            random_state=random_state
        )

        # Checking dataset size
        assert len(dataset) == n_sequences

        # Checking classes
        assert isinstance(dataset.y, np.ndarray)
        assert dataset.y.shape == (n_sequences,)
        assert set(dataset.y) == set(range(n_classes))

        # Checking sequence lengths and number of features
        if isinstance(length_range, tuple):
            a, b = length_range
            assert all((a <= len(x) <= b) and x.shape[1] == n_features for x in dataset.X)
        else:
            X = np.array(dataset.X)
            assert X.shape == (n_sequences, length_range, n_features)

# _sample_prior

@pytest.mark.parametrize(
    'length,n_features',
    list(product([5, 10, 15], [1, 2, 5]))
)
def test_sample_prior(length, n_features):
    y = _sample_prior(K, length, n_features, random_state)
    assert y.shape == (length, n_features)

# _K

def test_K():
    X = np.expand_dims(np.arange(5), axis=1)
    cov = K(X, X)
    assert_equal(cov, np.array([
        [1.00000000e+00, 6.06530660e-01, 1.35335283e-01, 1.11089965e-02, 3.35462628e-04],
        [6.06530660e-01, 1.00000000e+00, 6.06530660e-01, 1.35335283e-01, 1.11089965e-02],
        [1.35335283e-01, 6.06530660e-01, 1.00000000e+00, 6.06530660e-01, 1.35335283e-01],
        [1.11089965e-02, 1.35335283e-01, 6.06530660e-01, 1.00000000e+00, 6.06530660e-01],
        [3.35462628e-04, 1.11089965e-02, 1.35335283e-01, 6.06530660e-01, 1.00000000e+00]
    ]))


import numpy as np
from sklearn.model_selection import train_test_split

class Dataset:
    """Represents a generic dataset.

    Parameters
    ----------
    X: array-like
        Data instances.

    y: array-like
        Labels corresponding to data instances.

    classes: array-like
        The complete set of possible classes/labels.

    random_state: numpy.random.RandomState, int, optional
        A random state object or seed for reproducible randomness.
    """
    def __init__(self, X, y, classes, random_state=None):
        self.X = X
        self.y = y
        self.classes = classes
        self.random_state = random_state

    def __len__(self):
        return len(self.y)

    def __getitem__(self, i):
        return self.X[i], self.y[i]

    def __iter__(self):
        for i in range(len(self)):
            yield self[i]

    def data(self):
        """Fetch the instances and labels.

        Returns
        -------
        X: array-like
            Data instances.

        y: array-like
            Labels corresponding to data instances.
        """
        return self.X, self.y

    def iter_by_class(self):
        """Generator for iterating through instances partitioned by class.

        Returns
        -------
        instances: generator yielding ``(instances, class)``
            Instances belong to each class.
        """
        X_np = np.array(self.X, dtype=object)
        for c in self.classes:
            yield X_np[self.y == c].tolist(), c

    def split(self, split_size, stratify=True, shuffle=True):
        """Splits the dataset into two smaller :class:`Dataset` objects.

        Parameters
        ----------
        split_size: 0 < float < 1
            Proportion of instances to be allocated to the second split.

        stratify: bool
            Whether or not stratify the split on the labels such that each split
            has a similar distribution of labels.

        shuffle: bool
            Whether or not to shuffle the data before partitioniing it.

        Returns
        -------
        split_1: :class:`Dataset`
            First dataset split.

        split_2: :class:`Dataset`
            Second dataset split.
        """
        X1, X2, y1, y2 = train_test_split(
            self.X, self.y,
            test_size=split_size,
            random_state=self.random_state,
            shuffle=shuffle,
            stratify=(self.y if stratify else None)
        )
        return (
            Dataset(X1, y1, self.classes, self.random_state),
            Dataset(X2, y2, self.classes, self.random_state)
        )



import os, pickle, pytest, warnings, os, numpy as np, hmmlearn.hmm
from copy import deepcopy
from sequentia.classifiers import GMMHMM, HMMClassifier, _ErgodicTopology
from sequentia.datasets import load_random_sequences
from ....support import assert_equal, assert_not_equal

# pytest.skip('Skip until datasets module is added and positive definite issues are fixed', allow_module_level=True)

# Set seed for reproducible randomness
random_state = np.random.RandomState(0)

# Create some sample data
dataset = load_random_sequences(50, n_features=2, n_classes=5, length_range=(10, 30), random_state=random_state)
dataset.classes = [f'c{i}' for i in range(5)]
dataset.y = np.array([f'c{i}' for i in dataset.y])
x, y = dataset[0]

# Create and fit some HMMs
hmms = []
for sequences, label in dataset.iter_by_class():
    hmm = GMMHMM(label=label, n_states=5, random_state=random_state)
    hmm.set_random_initial()
    hmm.set_random_transitions()
    hmm.fit(sequences)
    hmms.append(hmm)

# Fit a classifier
hmm_clf = HMMClassifier()
hmm_clf.fit(hmms)

# =================== #
# HMMClassifier.fit() #
# =================== #

def test_fit_list():
    """Fit the classifier using a list of HMMs"""
    clf = HMMClassifier()
    clf.fit(hmms)
    assert clf.models_ == hmms

def test_fit_list_empty():
    """Fit the classifier using an empty list"""
    clf = HMMClassifier()
    with pytest.raises(RuntimeError) as e:
        clf.fit([])
    assert str(e.value) == 'The classifier must be fitted with at least one HMM'

def test_fit_list_invalid():
    """Fit the classifier using a list of invalid types"""
    clf = HMMClassifier()
    with pytest.raises(TypeError) as e:
        clf.fit(['a', 'b'])
    assert str(e.value) == 'Expected all models to be GMMHMM objects'

# ======================= #
# HMMClassifier.predict() #
# ======================= #

def test_predict_single_frequency_prior():
    """Predict a single observation sequence with a frequency prior"""
    prediction = hmm_clf.predict(x, prior='frequency', return_scores=False, original_labels=False)
    assert prediction == 1

def test_predict_single_uniform_prior():
    """Predict a single observation sequence with a uniform prior"""
    prediction = hmm_clf.predict(x, prior='uniform', return_scores=False, original_labels=False)
    assert prediction == 1

def test_predict_single_custom_prior():
    """Predict a single observation sequence with a custom prior"""
    prediction = hmm_clf.predict(x, prior=([1e-50]*4+[1-4e-50]), return_scores=False, original_labels=False)
    assert prediction == 1

def test_predict_single_return_scores():
    """Predict a single observation sequence and return the transformed label, with the un-normalized posterior scores"""
    prediction = hmm_clf.predict(x, prior='frequency', return_scores=True, original_labels=False)
    assert isinstance(prediction, tuple)
    assert prediction[0] == 1
    assert_equal(prediction[1], np.array([
        -131.46105165, -78.80931343, -99.35179093, -90.89464994, -483.92229446
    ]))

def test_predict_single_original_labels():
    """Predict a single observation sequence and return the original label, without the un-normalized posterior scores"""
    prediction = hmm_clf.predict(x, prior='uniform', return_scores=False, original_labels=True)
    assert prediction == 'c1'

def test_predict_single_return_scores_original_labels():
    """Predict a single observation sequence and return the original label, with the un-normalized posterior scores"""
    prediction = hmm_clf.predict(x, prior='frequency', return_scores=True, original_labels=True)
    assert isinstance(prediction, tuple)
    assert prediction[0] == 'c1'
    assert_equal(prediction[1], np.array([
        -131.46105165, -78.80931343, -99.35179093, -90.89464994, -483.92229446
    ]))

def test_predict_multiple_frequency_prior():
    """Predict multiple observation sequences with a frequency prior"""
    predictions = hmm_clf.predict(dataset.X, prior='frequency', return_scores=False, original_labels=False)
    assert_equal(predictions, np.array([
        1, 3, 1, 0, 1, 3, 1, 1, 3, 2, 1, 2, 3, 2, 4, 1, 3, 2, 0, 0, 1, 1,
        3, 1, 1, 3, 1, 1, 1, 1, 4, 3, 0, 3, 1, 2, 3, 1, 2, 1, 1, 1, 3, 3,
        2, 3, 1, 1, 4, 1
    ]))

def test_predict_multiple_uniform_prior():
    """Predict multiple observation sequences with a uniform prior"""
    predictions = hmm_clf.predict(dataset.X, prior='uniform', return_scores=False, original_labels=False)
    assert_equal(predictions, np.array([
        1, 3, 1, 0, 1, 3, 1, 1, 3, 2, 1, 2, 3, 2, 4, 1, 3, 2, 0, 0, 1, 1,
        3, 1, 1, 3, 1, 1, 1, 1, 4, 3, 0, 3, 1, 2, 3, 1, 2, 1, 1, 1, 3, 3,
        2, 3, 1, 1, 4, 1
    ]))

def test_predict_multiple_custom_prior():
    """Predict multiple observation sequences with a custom prior"""
    predictions = hmm_clf.predict(dataset.X, prior=([1-4e-50]+[1e-50]*4), return_scores=False, original_labels=False)
    assert_equal(predictions, np.array([
        0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 0, 0, 0, 0, 0,
        0, 0, 0, 0, 0, 0
    ]))

def test_predict_multiple_return_scores():
    """Predict multiple observation sequences and return the transformed labels, with the un-normalized posterior scores"""
    predictions = hmm_clf.predict(dataset.X, prior='frequency', return_scores=True, original_labels=False)
    assert isinstance(predictions, tuple)
    assert_equal(predictions[0], np.array([
        1, 3, 1, 0, 1, 3, 1, 1, 3, 2, 1, 2, 3, 2, 4, 1, 3, 2, 0, 0, 1, 1,
        3, 1, 1, 3, 1, 1, 1, 1, 4, 3, 0, 3, 1, 2, 3, 1, 2, 1, 1, 1, 3, 3,
        2, 3, 1, 1, 4, 1
    ]))
    assert_equal(predictions[1][:5], np.array([
        [-131.46105165,  -78.80931343,  -99.35179093,  -90.89464994, -483.92229446],
        [ -91.58935678,  -66.6556658 ,  -91.46883547,  -65.69934269, -716.797869  ],
        [ -97.5230626 ,  -74.50878143,  -99.1544397 ,  -76.48361176, -690.2988915 ],
        [  14.24986519,  -44.85298283,  -41.50143234,  -40.50844881, -148.67734234],
        [ -95.11368472,  -40.81069058,  -59.46841129,  -52.60034218, -430.36823963]
    ]))

def test_predict_multiple_original_labels():
    """Predict multiple observation sequences and return the original labels, without the un-normalized posterior scores"""
    predictions = hmm_clf.predict(dataset.X, prior='frequency', return_scores=False, original_labels=True)
    assert all(np.equal(
        predictions.astype(object),
        np.array([
            'c1', 'c3', 'c1', 'c0', 'c1', 'c3', 'c1', 'c1', 'c3', 'c2', 'c1',
            'c2', 'c3', 'c2', 'c4', 'c1', 'c3', 'c2', 'c0', 'c0', 'c1', 'c1',
            'c3', 'c1', 'c1', 'c3', 'c1', 'c1', 'c1', 'c1', 'c4', 'c3', 'c0',
            'c3', 'c1', 'c2', 'c3', 'c1', 'c2', 'c1', 'c1', 'c1', 'c3', 'c3',
            'c2', 'c3', 'c1', 'c1', 'c4', 'c1'
        ], dtype=object)
    ))

def test_predict_multiple_return_scores_original_labels():
    """Predict multiple observation sequences and return the original labels, with the un-normalized posterior scores"""
    predictions = hmm_clf.predict(dataset.X, prior='frequency', return_scores=True, original_labels=True)
    assert isinstance(predictions, tuple)
    assert all(np.equal(
        predictions[0].astype(object),
        np.array([
            'c1', 'c3', 'c1', 'c0', 'c1', 'c3', 'c1', 'c1', 'c3', 'c2', 'c1',
            'c2', 'c3', 'c2', 'c4', 'c1', 'c3', 'c2', 'c0', 'c0', 'c1', 'c1',
            'c3', 'c1', 'c1', 'c3', 'c1', 'c1', 'c1', 'c1', 'c4', 'c3', 'c0',
            'c3', 'c1', 'c2', 'c3', 'c1', 'c2', 'c1', 'c1', 'c1', 'c3', 'c3',
            'c2', 'c3', 'c1', 'c1', 'c4', 'c1'
        ], dtype=object)
    ))
    assert_equal(predictions[1][:5], np.array([
        [-131.46105165,  -78.80931343,  -99.35179093,  -90.89464994, -483.92229446],
        [ -91.58935678,  -66.6556658 ,  -91.46883547,  -65.69934269, -716.797869  ],
        [ -97.5230626 ,  -74.50878143,  -99.1544397 ,  -76.48361176, -690.2988915 ],
        [  14.24986519,  -44.85298283,  -41.50143234,  -40.50844881, -148.67734234],
        [ -95.11368472,  -40.81069058,  -59.46841129,  -52.60034218, -430.36823963]
    ]))

# ======================== #
# HMMClassifier.evaluate() #
# ======================== #

def test_evaluate():
    """Evaluate performance on some observation sequences and labels"""
    acc, cm = hmm_clf.evaluate(dataset.X, dataset.y, prior='frequency')
    assert acc == 0.92
    print(repr(cm))
    assert_equal(cm, np.array([
        [ 4,  0,  0,  0,  0],
        [ 0, 20,  0,  1,  0],
        [ 0,  1,  7,  0,  0],
        [ 0,  2,  0, 12,  0],
        [ 0,  0,  0,  0,  3]
    ]))

# ==================== #
# HMMClassifier.save() #
# ==================== #

def test_save_unfitted():
    """Save an unfitted HMMClassifier object."""
    try:
        with pytest.raises(AttributeError) as e:
            HMMClassifier().save('test.pkl')
        assert str(e.value) == 'No models available - the classifier must be fitted first'
    finally:
        if os.path.exists('test.pkl'):
            os.remove('test.pkl')

def test_save_fitted():
    """Save a fitted HMMClassifier object."""
    try:
        hmm_clf.save('test.pkl')
        assert os.path.isfile('test.pkl')
    finally:
        os.remove('test.pkl')

# ==================== #
# HMMClassifier.load() #
# ==================== #

def test_load_invalid_format():
    """Load a HMMClassifier from an illegally formatted file"""
    try:
        with open('test', 'w') as f:
            f.write('illegal')
        with pytest.raises(pickle.UnpicklingError) as e:
            HMMClassifier.load('test')
    finally:
        os.remove('test')

def test_load_valid():
    """Load a serialized HMMClassifier"""
    try:
        hmm_clf.save('test.pkl')
        clf = HMMClassifier.load('test.pkl')
        # Check that all fields are still the same
        assert isinstance(clf, HMMClassifier)
        assert all(isinstance(model, GMMHMM) for model in clf.models_)
        assert [model.label for model in clf.models_] == dataset.classes
        assert list(clf.encoder_.classes_) == dataset.classes
        predictions = clf.predict(dataset.X, prior='frequency', return_scores=True, original_labels=True)
        assert isinstance(predictions, tuple)
        assert all(np.equal(
            predictions[0].astype(object),
            np.array([
                'c1', 'c3', 'c1', 'c0', 'c1', 'c3', 'c1', 'c1', 'c3', 'c2', 'c1',
                'c2', 'c3', 'c2', 'c4', 'c1', 'c3', 'c2', 'c0', 'c0', 'c1', 'c1',
                'c3', 'c1', 'c1', 'c3', 'c1', 'c1', 'c1', 'c1', 'c4', 'c3', 'c0',
                'c3', 'c1', 'c2', 'c3', 'c1', 'c2', 'c1', 'c1', 'c1', 'c3', 'c3',
                'c2', 'c3', 'c1', 'c1', 'c4', 'c1'
            ], dtype=object)
        ))
        assert_equal(predictions[1][:5], np.array([
            [-131.46105165,  -78.80931343,  -99.35179093,  -90.89464994, -483.92229446],
            [ -91.58935678,  -66.6556658 ,  -91.46883547,  -65.69934269, -716.797869  ],
            [ -97.5230626 ,  -74.50878143,  -99.1544397 ,  -76.48361176, -690.2988915 ],
            [  14.24986519,  -44.85298283,  -41.50143234,  -40.50844881, -148.67734234],
            [ -95.11368472,  -40.81069058,  -59.46841129,  -52.60034218, -430.36823963]
        ]))
    finally:
        os.remove('test.pkl')


import numpy as np
from sequentia.datasets import Dataset
from ...support import assert_all_equal

# Set seed for reproducible randomness
seed = 0
np.random.seed(seed)
rng = np.random.RandomState(seed)

n_sequences = 100
n_classes = 5
classes = range(n_classes)

X = [np.array([[i]]) for i in range(100)]
y = rng.choice(classes, n_sequences)

dataset = Dataset(X=X, y=y, classes=classes, random_state=rng)

def test_dataset_iter_by_class():
    true_partitions = [[x for (x, label) in zip(X, y) if label == c] for c in classes]
    generated_partitions = [sequences for sequences, _ in dataset.iter_by_class()]
    assert_all_equal(true_partitions, generated_partitions)

def test_dataset_split():
    train_set, test_set = dataset.split(split_size=0.2)
    assert len(train_set) == 80
    assert len(test_set) == 20
    assert sorted((*train_set.y, *test_set.y)) == sorted(y)



import numpy as np, scipy.spatial
from tslearn.utils import to_time_series_dataset
from tslearn.clustering import TimeSeriesKMeans
from .base import Dataset
from ..internals import _Validator

def load_random_sequences(
    n_sequences, n_features, n_classes, length_range,
    variance_range=(2, 5), lengthscale_range=(0.2, 0.5),
    random_state=None, tslearn_kwargs={}
):
    """Generates random sequences by sampling from a Gaussian process prior and clustering sequences to obtained labels.

    **Generating sequences**

    The GP prior is parameterised by a kernel function :math:`k_\\theta(x, x')`.
    In this case, we use the squared exponential kernel which is parameterised by :math:`\\theta=(\sigma^2, l)`.

    .. math::
        k_\\theta(x, x') = \\sigma^2 \\exp \\left( - \\frac{(x - x')^2}{2l^2} \\right)

    - :math:`\\sigma^2` is the **variance** of the kernel, which controls the height of the generated functions.
    - :math:`l` is the **lengthscale** of the kernel, which controls the distance between the peaks and troughs of the generated functions.

    For :math:`x` values :math:`\mathbf{x}=(0,\\ldots, n-1)` where :math:`n` is the specified length of the generated sequence, we compute a
    kernel matrix where :math:`K_\\theta(\mathbf{x}, \mathbf{x})_{ij}=k_\\theta(x_i, x_j)`.

    We obtain function values over :math:`X` by drawing samples from a multivariate normal distribution.

    .. math::
        \\mathbf{y} \\sim \\mathcal{N} \\big( \mathbf{0}, K_\\theta(\mathbf{x}, \mathbf{x}) \\big)

    Each sequence is drawn from its own independent GP prior, with a variance and lengthscale sampled from uniform hyperpriors.

    .. math::
        \\sigma^2 \\sim U(a_{\\sigma^2},b_{\\sigma^2}) \qquad l \\sim U(a_l,b_l)

    Where :math:`(a_{\\sigma^2},b_{\\sigma^2})` and :math:`(a_l,b_l)` are given by ``variance_range`` and ``lengthscale_range``.

    If ``n_features`` is more than one, then the same :math:`\\sigma^2` and :math:`l` values are shared along each feature of a single
    sequence, which means that features are often very highly correlated.

    **Generating labels**

    In order to generate labels that are assigned to generated sequences that are somewhat similar,
    we perform clustering over these sequences using :class:`tslearn.clustering.TimeSeriesKMeans`
    and use the resulting cluster labels as the label for each sequence.

    Parameters
    ----------
    n_sequences: int
        Number of sequences to generate.

    n_features: int
        Number of features in each sequences.

    n_classes: int
        Number of classes.

    length_range: tuple(int, int) or int
        Range of values to uniformly sample sequence lengths from.

        If a single value is specified, then all sequences will have equal length.

    variance_range: tuple(float, float) or float
        Lower and upper range of the uniform hyperprior on the GP prior's variance kernel hyperparameter.

        If a single value is specified, then this value is used as the variance rather than placing a hyperprior.

    lengthscale_range: tuple(float, float) or float
        Lower and upper range of the uniform hyperprior on the GP prior's lengthscale kernel hyperparameter.

        If a single value is specified, then this value is used as the lengthscale rather than placing a hyperprior.

    random_state: numpy.random.RandomState, int, optional
        A random state object or seed for reproducible randomness.

    tslearn_kwargs: dict
        Additional key-word arguments for the :class:`tslearn.clustering.TimeSeriesKMeans` constructor.

        The defaults are:

        - ``'metric'``: ``'dtw'``
        - ``'max_iter'``: ``5``
        - ``'max_iter_barycenter'``: ``5``
        - ``'n_jobs'``: ``-1``

        ``'n_clusters'`` and ``'random_state'`` are overwritten by the ``n_classes`` and ``random_state`` arguments
        given to this function.

    Returns
    -------
    dataset: :class:`sequentia.datasets.Dataset`
        A dataset object representing the loaded digits.
    """
    val = _Validator()
    random_state = val.is_random_state(random_state)
    n_classes = val.is_restricted_integer(
        n_classes, lambda x: x <= n_sequences,
        desc='number of classes', expected='no more than n_sequences'
    )

    # Set default tslearn key-word arguments
    tslearn_kwargs['metric'] = tslearn_kwargs.get('metric', 'dtw')
    tslearn_kwargs['max_iter'] = tslearn_kwargs.get('max_iter', 5)
    tslearn_kwargs['max_iter_barycenter'] = tslearn_kwargs.get('max_iter_barycenter', 5)
    tslearn_kwargs['n_jobs'] = tslearn_kwargs.get('n_jobs', -1)

    # Override tslearn key-word arguments
    tslearn_kwargs['n_clusters'] = n_classes
    tslearn_kwargs['random_state'] = random_state

    # Sample sequences from GP priors
    X = []
    for i in range(n_sequences):
        length = _sample_from_range(length_range, int, random_state=random_state)
        variance = _sample_from_range(variance_range, float, random_state=random_state)
        lengthscale = _sample_from_range(lengthscale_range, float, random_state=random_state)
        k = lambda x1, x2: _K(x1, x2, variance=variance, lengthscale=lengthscale)
        X.append(_sample_prior(k, length, n_features, random_state))

    if n_sequences == 1:
        y = np.array([0])
    elif n_classes == 1:
        y = np.zeros(n_sequences)
    else:
        # Cluster sequences to obtain labels
        y = TimeSeriesKMeans(**tslearn_kwargs).fit_predict(to_time_series_dataset(X))

    return Dataset(X, y, range(n_classes), random_state)

def _sample_prior(k, length, n_features, random_state):
    X = np.expand_dims(np.linspace(0, length, length), axis=1)
    return random_state.multivariate_normal(np.zeros(length), k(X, X), size=n_features).T

def _K(x1, x2, variance, lengthscale):
    return variance * np.exp(-scipy.spatial.distance.cdist(x1, x2, 'sqeuclidean') / 2*lengthscale**2)

def _sample_from_range(range_var, var_type, random_state):
    if isinstance(range_var, (int, float)):
        return range_var
    if var_type == int:
        return random_state.choice(range(range_var[0], range_var[1] + 1))
    elif var_type == float:
        return random_state.uniform(*range_var)



import pytest, warnings, os, math, numpy as np, hmmlearn.base, hmmlearn.hmm
from copy import deepcopy
from sequentia.classifiers import GMMHMM, _LeftRightTopology, _ErgodicTopology, _LinearTopology
from sequentia.datasets import load_random_sequences
from ....support import assert_equal, assert_not_equal, assert_all_equal, assert_all_not_equal

# Set seed for reproducible randomness
random_state = np.random.RandomState(0)

# Create some sample data
dataset = load_random_sequences(15, n_features=2, n_classes=2, length_range=(20, 30), random_state=random_state)
X = [x for x, y in dataset if y == 0]
x = X[0]

# Unparameterized HMMs
hmm_lr = GMMHMM(label='c1', n_states=5, topology='left-right', random_state=random_state)
hmm_e = GMMHMM(label='c1', n_states=5, topology='ergodic', random_state=random_state)
hmm_lin = GMMHMM(label='c1', n_states=5, topology='linear', random_state=random_state)

# ========================================================= #
# GMMHMM.set_uniform_initial() + GMMHMM.initial_ (property) #
# ========================================================= #

def test_left_right_uniform_initial():
    """Uniform initial state distribution for a left-right HMM"""
    hmm = deepcopy(hmm_lr)
    hmm.set_uniform_initial()
    assert_equal(hmm.initial_, np.array([
        0.2, 0.2, 0.2, 0.2, 0.2
    ]))

def test_ergodic_uniform_initial():
    """Uniform initial state distribution for an ergodic HMM"""
    hmm = deepcopy(hmm_e)
    hmm.set_uniform_initial()
    assert_equal(hmm.initial_, np.array([
        0.2, 0.2, 0.2, 0.2, 0.2
    ]))

def test_linear_uniform_initial():
    """Uniform initial state distribution for a linear HMM"""
    hmm = deepcopy(hmm_lin)
    hmm.set_uniform_initial()
    assert_equal(hmm.initial_, np.array([
        0.2, 0.2, 0.2, 0.2, 0.2
    ]))

# ======================================================== #
# GMMHMM.set_random_initial() + GMMHMM.initial_ (property) #
# ======================================================== #

def test_left_right_random_initial():
    """Random initial state distribution for a left-right HMM"""
    hmm = deepcopy(hmm_lr)
    hmm.set_random_initial()
    assert_equal(hmm.initial_, np.array([
        0.53803322, 0.12404781, 0.07762362, 0.17663443, 0.08366092
    ]))

def test_ergodic_random_initial():
    """Random initial state distribution for an ergodic HMM"""
    hmm = deepcopy(hmm_e)
    hmm.set_random_initial()
    assert_equal(hmm.initial_, np.array([
        0.53803322, 0.12404781, 0.07762362, 0.17663443, 0.08366092
    ]))

def test_linear_random_initial():
    """Random initial state distribution for a linear HMM"""
    hmm = deepcopy(hmm_lin)
    hmm.set_random_initial()
    assert_equal(hmm.initial_, np.array([
        0.53803322, 0.12404781, 0.07762362, 0.17663443, 0.08366092
    ]))

# ================================================================= #
# GMMHMM.set_uniform_transitions() + GMMHMM.transitions_ (property) #
# ================================================================= #

def test_left_right_uniform_transitions():
    """Uniform transition matrix for a left-right HMM"""
    hmm = deepcopy(hmm_lr)
    hmm.set_uniform_transitions()
    assert_equal(hmm.transitions_, np.array([
        [0.2, 0.2 , 0.2       , 0.2       , 0.2       ],
        [0. , 0.25, 0.25      , 0.25      , 0.25      ],
        [0. , 0.  , 0.33333333, 0.33333333, 0.33333333],
        [0. , 0.  , 0.        , 0.5       , 0.5       ],
        [0. , 0.  , 0.        , 0.        , 1.        ]
    ]))

def test_ergodic_uniform_transitions():
    """Uniform transition matrix for an ergodic HMM"""
    hmm = deepcopy(hmm_e)
    hmm.set_uniform_transitions()
    assert_equal(hmm.transitions_, np.array([
        [0.2, 0.2, 0.2, 0.2, 0.2],
        [0.2, 0.2, 0.2, 0.2, 0.2],
        [0.2, 0.2, 0.2, 0.2, 0.2],
        [0.2, 0.2, 0.2, 0.2, 0.2],
        [0.2, 0.2, 0.2, 0.2, 0.2]
    ]))

def test_linear_uniform_transitions():
    """Uniform transition matrix for a linear HMM"""
    hmm = deepcopy(hmm_lin)
    hmm.set_uniform_transitions()
    assert_equal(hmm.transitions_, np.array([
        [0.5, 0.5, 0. , 0. , 0. ],
        [0. , 0.5, 0.5, 0. , 0. ],
        [0. , 0. , 0.5, 0.5, 0. ],
        [0. , 0. , 0. , 0.5, 0.5],
        [0. , 0. , 0. , 0. , 1. ]
    ]))

# ================================================================ #
# GMMHMM.set_random_transitions() + GMMHMM.transitions_ (property) #
# ================================================================ #

def test_left_right_random_transitions():
    """Random transition matrix for a left-right HMM"""
    hmm = deepcopy(hmm_lr)
    hmm.set_random_transitions()
    assert_equal(hmm.transitions_, np.array([
        [0.53803322, 0.12404781, 0.07762362, 0.17663443, 0.08366092],
        [0.        , 0.0544546 , 0.167254  , 0.43679272, 0.34149867],
        [0.        , 0.        , 0.02569653, 0.93686415, 0.03743932],
        [0.        , 0.        , 0.        , 0.80245882, 0.19754118],
        [0.        , 0.        , 0.        , 0.        , 1.        ]
    ]))

def test_ergodic_random_transitions():
    """Random transition matrix for an ergodic HMM"""
    hmm = deepcopy(hmm_e)
    hmm.set_random_transitions()
    assert_equal(hmm.transitions_, np.array([
        [0.53803322, 0.12404781, 0.07762362, 0.17663443, 0.08366092],
        [0.05407134, 0.16607684, 0.43371851, 0.33909515, 0.00703816],
        [0.12935118, 0.00516918, 0.67173292, 0.16536041, 0.02838631],
        [0.15768347, 0.31907791, 0.42873228, 0.06083948, 0.03366686],
        [0.42607069, 0.17697038, 0.33288653, 0.04212738, 0.02194502]
    ]))

def test_linear_random_transitions():
    """Random transition matrix for a linear HMM"""
    hmm = deepcopy(hmm_lin)
    hmm.set_random_transitions()
    assert_equal(hmm.transitions_, np.array([
        [0.81263954, 0.18736046, 0.        , 0.        , 0.        ],
        [0.        , 0.30529464, 0.69470536, 0.        , 0.        ],
        [0.        , 0.        , 0.34435856, 0.65564144, 0.        ],
        [0.        , 0.        , 0.        , 0.27688918, 0.72311082],
        [0.        , 0.        , 0.        , 0.        , 1.        ]
    ]))

# ============================= #
# GMMHMM.fit() + GMMHMM.n_seqs_ #
# ============================= #

def test_fit_without_initial_and_transition():
    """Fitting before setting the initial state distribution and transition matrix"""
    hmm = deepcopy(hmm_lr)
    with pytest.raises(AttributeError) as e:
        hmm.fit(X)
    assert str(e.value) == 'No initial state distribution has been defined'

def test_fit_without_transitions():
    """Fitting before setting the initial transition matrix"""
    hmm = deepcopy(hmm_lr)
    hmm.set_uniform_initial()
    with pytest.raises(AttributeError) as e:
        hmm.fit(X)
    assert str(e.value) == 'No transition matrix has been defined'

def test_fit_without_initial():
    """Fitting before setting the initial state distribution"""
    hmm = deepcopy(hmm_lr)
    hmm.set_uniform_transitions()
    with pytest.raises(AttributeError) as e:
        hmm.fit(X)
    assert str(e.value) == 'No initial state distribution has been defined'

def test_fit_sets_internals():
    """Check that fitting sets internal attributes"""
    hmm = deepcopy(hmm_lr)
    hmm.set_uniform_initial()
    hmm.set_uniform_transitions()
    hmm.fit(X)
    assert hmm.n_seqs_ == 8
    assert isinstance(hmm.model, hmmlearn.hmm.GMMHMM)

def test_left_right_fit_updates_uniform_initial():
    """Check fitting with a uniform initial state distribution of a left-right HMM"""
    hmm = deepcopy(hmm_lr)
    hmm.set_uniform_initial()
    hmm.set_uniform_transitions()
    before = hmm.initial_
    assert_equal(before, np.array([
        0.2, 0.2, 0.2, 0.2, 0.2
    ]))
    hmm.fit(X)
    assert_not_equal(before, hmm.initial_)

def test_left_right_fit_updates_random_initial():
    """Check fitting with a random initial state distribution of a left-right HMM"""
    hmm = deepcopy(hmm_lr)
    hmm.set_random_initial()
    hmm.set_random_transitions()
    before = hmm.initial_
    assert_equal(before, np.array([
        0.53803322, 0.12404781, 0.07762362, 0.17663443, 0.08366092
    ]))
    hmm.fit(X)
    assert_not_equal(before, hmm.initial_)

def test_left_right_fit_updates_uniform_transitions():
    """Check fitting with a uniform transition matrix of a left-right HMM"""
    hmm = deepcopy(hmm_lr)
    hmm.set_uniform_initial()
    hmm.set_uniform_transitions()
    before = hmm.transitions_
    assert_equal(before, np.array([
        [0.2       , 0.2       , 0.2       , 0.2       , 0.2       ],
        [0.        , 0.25      , 0.25      , 0.25      , 0.25      ],
        [0.        , 0.        , 0.33333333, 0.33333333, 0.33333333],
        [0.        , 0.        , 0.        , 0.5       , 0.5       ],
        [0.        , 0.        , 0.        , 0.        , 1.        ]
    ]))
    hmm.fit(X)
    assert_not_equal(before, hmm.transitions_)

def test_left_right_fit_updates_random_transitions():
    """Check fitting with a random transition matrix of a left-right HMM"""
    hmm = deepcopy(hmm_lr)
    hmm.set_random_initial()
    hmm.set_random_transitions()
    before = hmm.transitions_
    assert_equal(before, np.array([
        [0.05407134, 0.16607684, 0.43371851, 0.33909515, 0.00703816],
        [0.        , 0.13313025, 0.0053202 , 0.69135803, 0.17019152],
        [0.        , 0.        , 0.11443295, 0.29289135, 0.59267569],
        [0.        , 0.        , 0.        , 0.87572918, 0.12427082],
        [0.        , 0.        , 0.        , 0.        , 1.        ]
    ]))
    hmm.fit(X)
    assert_not_equal(before, hmm.transitions_)

def test_ergodic_fit_updates_uniform_initial():
    """Check fitting with a uniform initial state distribution of an ergodic HMM"""
    hmm = deepcopy(hmm_e)
    hmm.set_uniform_initial()
    hmm.set_uniform_transitions()
    before = hmm.initial_
    assert_equal(before, np.array([
        0.2, 0.2, 0.2, 0.2, 0.2
    ]))
    hmm.fit(X)
    assert_not_equal(before, hmm.initial_)

def test_ergodic_fit_updates_random_initial():
    """Check fitting with a random initial state distribution of an ergodic HMM"""
    hmm = deepcopy(hmm_e)
    hmm.set_random_initial()
    hmm.set_random_transitions()
    before = hmm.initial_
    assert_equal(before, np.array([
        0.53803322, 0.12404781, 0.07762362, 0.17663443, 0.08366092
    ]))
    hmm.fit(X)
    assert_not_equal(before, hmm.initial_)

def test_ergodic_fit_updates_uniform_transitions():
    """Check fitting with a uniform transition matrix of an ergodic HMM"""
    hmm = deepcopy(hmm_e)
    hmm.set_uniform_initial()
    hmm.set_uniform_transitions()
    before = hmm.transitions_
    assert_equal(before, np.array([
        [0.2, 0.2, 0.2, 0.2, 0.2],
        [0.2, 0.2, 0.2, 0.2, 0.2],
        [0.2, 0.2, 0.2, 0.2, 0.2],
        [0.2, 0.2, 0.2, 0.2, 0.2],
        [0.2, 0.2, 0.2, 0.2, 0.2]
    ]))
    hmm.fit(X)
    assert_not_equal(before, hmm.transitions_)

def test_ergodic_fit_updates_random_transitions():
    """Check fitting with a random transition matrix of an ergodic HMM"""
    hmm = deepcopy(hmm_e)
    hmm.set_random_initial()
    hmm.set_random_transitions()
    before = hmm.transitions_
    assert_equal(before, np.array([
        [0.05407134, 0.16607684, 0.43371851, 0.33909515, 0.00703816],
        [0.12935118, 0.00516918, 0.67173292, 0.16536041, 0.02838631],
        [0.15768347, 0.31907791, 0.42873228, 0.06083948, 0.03366686],
        [0.42607069, 0.17697038, 0.33288653, 0.04212738, 0.02194502],
        [0.20328414, 0.13729798, 0.03560389, 0.4874536 , 0.13636039]
    ]))
    hmm.fit(X)
    assert_not_equal(before, hmm.transitions_)

def test_linear_fit_updates_uniform_initial():
    """Check fitting with a uniform initial state distribution of a linear HMM"""
    hmm = deepcopy(hmm_lin)
    hmm.set_uniform_initial()
    hmm.set_uniform_transitions()
    before = hmm.initial_
    assert_equal(before, np.array([
        0.2, 0.2, 0.2, 0.2, 0.2
    ]))
    hmm.fit(X)
    assert_not_equal(before, hmm.initial_)

def test_linear_fit_updates_random_initial():
    """Check fitting with a random initial state distribution of a linear HMM"""
    hmm = deepcopy(hmm_lin)
    hmm.set_random_initial()
    hmm.set_random_transitions()
    before = hmm.initial_
    assert_equal(before, np.array([
        0.53803322, 0.12404781, 0.07762362, 0.17663443, 0.08366092
    ]))
    hmm.fit(X)
    assert_not_equal(before, hmm.initial_)

def test_linear_fit_updates_uniform_transitions():
    """Check fitting with a uniform transition matrix of a linear HMM"""
    hmm = deepcopy(hmm_lin)
    hmm.set_uniform_initial()
    hmm.set_uniform_transitions()
    before = hmm.transitions_
    assert_equal(before, np.array([
        [0.5, 0.5, 0. , 0. , 0. ],
        [0. , 0.5, 0.5, 0. , 0. ],
        [0. , 0. , 0.5, 0.5, 0. ],
        [0. , 0. , 0. , 0.5, 0.5],
        [0. , 0. , 0. , 0. , 1. ]
    ]))
    hmm.fit(X)
    assert_not_equal(before, hmm.transitions_)

def test_linear_fit_updates_random_transitions():
    """Check fitting with a random transition matrix of a linear HMM"""
    hmm = deepcopy(hmm_lin)
    hmm.set_random_initial()
    hmm.set_random_transitions()
    before = hmm.transitions_
    print(repr(before))
    assert_equal(before, np.array([
        [0.24561338, 0.75438662, 0.        , 0.        , 0.        ],
        [0.        , 0.56122003, 0.43877997, 0.        , 0.        ],
        [0.        , 0.        , 0.02669601, 0.97330399, 0.        ],
        [0.        , 0.        , 0.        , 0.00763653, 0.99236347],
        [0.        , 0.        , 0.        , 0.        , 1.        ]
    ]))
    hmm.fit(X)
    assert_not_equal(before, hmm.transitions_)

# ================ #
# GMMHMM.forward() #
# ================ #

def test_forward_without_fit():
    """Forward algorithm without fitting the HMM"""
    hmm = deepcopy(hmm_lr)
    hmm.set_random_initial()
    hmm.set_random_transitions()
    with pytest.raises(AttributeError) as e:
        hmm.forward(x)
    assert str(e.value) == 'The model must be fitted first'

def test_left_right_forward():
    """Forward algorithm on a left-right HMM"""
    hmm = deepcopy(hmm_lr)
    hmm.set_random_initial()
    hmm.set_random_transitions()
    hmm.fit(X)
    assert math.isclose(hmm.forward(x), -89.59052551245605, rel_tol=1e-8)

def test_ergodic_forward():
    """Forward algorithm on an ergodic HMM"""
    hmm = deepcopy(hmm_e)
    hmm.set_random_initial()
    hmm.set_random_transitions()
    hmm.fit(X)
    assert math.isclose(hmm.forward(x), -97.67911812603418, rel_tol=1e-8)

def test_linear_forward():
    """Forward algorithm on a linear HMM"""
    hmm = deepcopy(hmm_lin)
    hmm.set_random_initial()
    hmm.set_random_transitions()
    hmm.fit(X)
    assert math.isclose(hmm.forward(x), -90.25666060143605, rel_tol=1e-8)

# =============== #
# GMMHMM.freeze() #
# =============== #

def test_freeze_no_params():
    hmm = deepcopy(hmm_lin)
    hmm.set_random_initial()
    hmm.set_random_transitions()
    hmm.freeze()
    assert hmm.frozen == set('stmcw')
    before = (hmm.initial_, hmm.transitions_)
    hmm.fit(X)
    assert_all_equal(before, (hmm.initial_, hmm.transitions_))

def test_freeze_invalid_params_type():
    hmm = deepcopy(hmm_lin)
    hmm.set_random_initial()
    hmm.set_random_transitions()
    with pytest.raises(TypeError) as e:
        hmm.freeze(0)
    assert str(e.value) == "Expected a string consisting of any combination of 's', 't', 'm', 'c', 'w'"

def test_freeze_invalid_params_content():
    hmm = deepcopy(hmm_lin)
    hmm.set_random_initial()
    hmm.set_random_transitions()
    with pytest.raises(ValueError) as e:
        hmm.freeze('stmcwz')
    assert str(e.value) == "Expected a string consisting of any combination of 's', 't', 'm', 'c', 'w'"

def test_freeze_valid_params():
    hmm = deepcopy(hmm_lin)
    hmm.set_random_initial()
    hmm.set_random_transitions()
    hmm.freeze('sw')
    assert hmm.frozen == set('sw')
    initial_before, transitions_before = hmm.initial_, hmm.transitions_
    hmm.fit(X)
    assert_equal(initial_before, hmm.initial_)
    assert_not_equal(transitions_before, hmm.transitions_)

def test_freeze_duplicate_params():
    hmm = deepcopy(hmm_lin)
    hmm.set_random_initial()
    hmm.set_random_transitions()
    hmm.freeze('swsswwssswww')
    assert hmm.frozen == set('sw')
    initial_before, transitions_before = hmm.initial_, hmm.transitions_
    hmm.fit(X)
    assert_equal(initial_before, hmm.initial_)
    assert_not_equal(transitions_before, hmm.transitions_)

def test_freeze_all_params():
    hmm = deepcopy(hmm_lin)
    hmm.set_random_initial()
    hmm.set_random_transitions()
    hmm.freeze('stmcw')
    assert hmm.frozen == set('stmcw')
    before = (hmm.initial_, hmm.transitions_)
    hmm.fit(X)
    assert_all_equal(before, (hmm.initial_, hmm.transitions_))

# ================= #
# GMMHMM.unfreeze() #
# ================= #

def test_unfreeze_no_params():
    hmm = deepcopy(hmm_lin)
    hmm.set_random_initial()
    hmm.set_random_transitions()
    hmm.freeze()
    assert hmm.frozen == set('stmcw')
    hmm.unfreeze()
    assert hmm.frozen == set()
    before = (hmm.initial_, hmm.transitions_)
    hmm.fit(X)
    assert_all_not_equal(before, (hmm.initial_, hmm.transitions_))

def test_unfreeze_invalid_params_type():
    hmm = deepcopy(hmm_lin)
    hmm.set_random_initial()
    hmm.set_random_transitions()
    hmm.freeze()
    with pytest.raises(TypeError) as e:
        hmm.unfreeze(0)
    assert str(e.value) == "Expected a string consisting of any combination of 's', 't', 'm', 'c', 'w'"

def test_unfreeze_invalid_params_content():
    hmm = deepcopy(hmm_lin)
    hmm.set_random_initial()
    hmm.set_random_transitions()
    hmm.freeze()
    with pytest.raises(ValueError) as e:
        hmm.unfreeze('stmcwz')
    assert str(e.value) == "Expected a string consisting of any combination of 's', 't', 'm', 'c', 'w'"

def test_unfreeze_valid_params():
    hmm = deepcopy(hmm_lin)
    hmm.set_random_initial()
    hmm.set_random_transitions()
    hmm.freeze('st')
    hmm.unfreeze('t')
    assert hmm.frozen == set('s')
    initial_before, transitions_before = hmm.initial_, hmm.transitions_
    hmm.fit(X)
    assert_equal(initial_before, hmm.initial_)
    assert_not_equal(transitions_before, hmm.transitions_)

def test_unfreeze_duplicate_params():
    hmm = deepcopy(hmm_lin)
    hmm.set_random_initial()
    hmm.set_random_transitions()
    hmm.freeze('st')
    hmm.unfreeze('tttmwcmwct')
    assert hmm.frozen == set('s')
    initial_before, transitions_before = hmm.initial_, hmm.transitions_
    hmm.fit(X)
    assert_equal(initial_before, hmm.initial_)
    assert_not_equal(transitions_before, hmm.transitions_)

def test_unfreeze_all_params():
    hmm = deepcopy(hmm_lin)
    hmm.set_random_initial()
    hmm.set_random_transitions()
    hmm.freeze()
    assert hmm.frozen == set('stmcw')
    hmm.unfreeze()
    assert hmm.frozen == set()
    before = (hmm.initial_, hmm.transitions_)
    hmm.fit(X)
    assert_all_not_equal(before, (hmm.initial_, hmm.transitions_))

# ======================= #
# GMMHMM.label (property) #
# ======================= #

def test_label():
    assert deepcopy(hmm_lr).label == 'c1'

# ========================== #
# GMMHMM.n_states (property) #
# ========================== #

def test_n_states():
    assert deepcopy(hmm_lr).n_states == 5

# ============================== #
# GMMHMM.n_components (property) #
# ============================== #

def test_n_components():
    assert deepcopy(hmm_lr).n_components == 1

# ================================= #
# GMMHMM.covariance_type (property) #
# ================================= #

def test_covariance_type():
    assert deepcopy(hmm_lr).covariance_type == 'full'

# ========================= #
# GMMHMM.n_seqs_ (property) #
# ========================= #

def test_n_seqs_without_fit():
    """Number of sequences without fitting the HMM"""
    hmm = deepcopy(hmm_lr)
    hmm.set_random_initial()
    hmm.set_random_transitions()
    with pytest.raises(AttributeError) as e:
        hmm.n_seqs_
    assert str(e.value) == 'The model has not been fitted and has not seen any observation sequences'

def test_n_seqs_with_fit():
    """Number of sequences after fitting the HMM"""
    hmm = deepcopy(hmm_lr)
    hmm.set_random_initial()
    hmm.set_random_transitions()
    hmm.fit(X)
    assert hmm.n_seqs_ == 8

# ======================== #
# GMMHMM.frozen (property) #
# ======================== #

def test_frozen():
    hmm = deepcopy(hmm_lr)
    hmm.freeze('sw')
    assert hmm.frozen == set('sw')

# ========================== #
# GMMHMM.monitor_ (property) #
# ========================== #

def test_monitor_without_fit():
    """Convergence monitor without fitting the HMM"""
    hmm = deepcopy(hmm_lr)
    hmm.set_random_initial()
    hmm.set_random_transitions()
    with pytest.raises(AttributeError) as e:
        hmm.monitor_
    assert str(e.value) == 'The model must be fitted first'

def test_monitor_with_fit():
    """Convergence monitor after fitting the HMM"""
    hmm = deepcopy(hmm_lr)
    hmm.set_random_initial()
    hmm.set_random_transitions()
    hmm.fit(X)
    assert isinstance(hmm.monitor_, hmmlearn.base.ConvergenceMonitor)

# ========================== #
# GMMHMM.weights_ (property) #
# ========================== #

def test_weights_without_fit():
    """GMM weights without fitting the HMM"""
    hmm = deepcopy(hmm_lr)
    hmm.set_random_initial()
    hmm.set_random_transitions()
    with pytest.raises(AttributeError) as e:
        hmm.weights_
    assert str(e.value) == 'The model must be fitted first'

def test_weights_with_fit():
    """GMM weights after fitting the HMM"""
    hmm = deepcopy(hmm_lr)
    hmm.set_random_initial()
    hmm.set_random_transitions()
    hmm.fit(X)
    assert_equal(hmm.weights_, np.array([
        [1.],
        [1.],
        [1.],
        [1.],
        [1.]
    ]))

# ======================== #
# GMMHMM.means_ (property) #
# ======================== #

def test_means_without_fit():
    """GMM mean vectors without fitting the HMM"""
    hmm = deepcopy(hmm_lr)
    hmm.set_random_initial()
    hmm.set_random_transitions()
    with pytest.raises(AttributeError) as e:
        hmm.means_
    assert str(e.value) == 'The model must be fitted first'

def test_means_with_fit():
    """GMM mean vectors after fitting the HMM"""
    hmm = deepcopy(hmm_lr)
    hmm.set_random_initial()
    hmm.set_random_transitions()
    hmm.fit(X)
    assert_equal(hmm.means_, np.array([
        [[ 0.49517361,  0.79670013]],
        [[ 1.81277369, -2.45995611]],
        [[-0.61198527, -0.2621587 ]],
        [[ 1.40168717,  0.16718235]],
        [[-2.05338535, -3.13926956]]
    ]))

# ========================= #
# GMMHMM.covars_ (property) #
# ========================= #

def test_covars_without_fit():
    """GMM covariance matrices without fitting the HMM"""
    hmm = deepcopy(hmm_lr)
    hmm.set_random_initial()
    hmm.set_random_transitions()
    with pytest.raises(AttributeError) as e:
        hmm.covars_
    assert str(e.value) == 'The model must be fitted first'

def test_covars_with_fit():
    """GMM covariance matrices after fitting the HMM"""
    hmm = deepcopy(hmm_lr)
    hmm.set_random_initial()
    hmm.set_random_transitions()
    hmm.fit(X)
    assert_equal(hmm.covars_, np.array([
        [[[ 1.38488559,  0.38570541],
          [ 0.38570541,  0.63293189]]],
        [[[ 1.73706667,  0.28568952],
          [ 0.28568952,  0.47176263]]],
        [[[ 0.99011246, -0.10938155],
          [-0.10938155,  0.01847633]]],
        [[[ 1.94877336, -0.17877102],
          [-0.17877102,  2.42880862]]],
        [[[ 3.52820275,  0.5571457 ],
          [ 0.5571457 ,  0.2716629 ]]]
    ]))

# ========================== #
# GMMHMM.initial_ (property) #
# ========================== #

def test_initial_without_setting():
    """Get initial state distribution without setting it"""
    hmm = deepcopy(hmm_lr)
    with pytest.raises(AttributeError) as e:
        hmm.initial_
    assert str(e.value) == 'No initial state distribution has been defined'

# ============================== #
# GMMHMM.transitions_ (property) #
# ============================== #

def test_transitions_without_setting():
    """Get transition matrix without setting it"""
    hmm = deepcopy(hmm_lr)
    with pytest.raises(AttributeError) as e:
        hmm.transitions_
    assert str(e.value) == 'No transition matrix has been defined'

# ======================== #
# GMMHMM.initial_ (setter) #
# ======================== #

def test_left_right_initial_left_right():
    """Set an initial state distribution generated by a left-right topology on a left-right HMM"""
    hmm = deepcopy(hmm_lr)
    topology = _LeftRightTopology(n_states=5, random_state=random_state)
    initial = topology.random_initial()
    hmm.initial_ = initial
    assert_equal(hmm.initial_, initial)

def test_left_right_initial_ergodic():
    """Set an initial state distribution generated by a left-right topology on an ergodic HMM"""
    hmm = deepcopy(hmm_lr)
    topology = _ErgodicTopology(n_states=5, random_state=random_state)
    initial = topology.random_initial()
    hmm.initial_ = initial
    assert_equal(hmm.initial_, initial)

def test_left_right_initial_linear():
    """Set an initial state distribution generated by a left-right topology on an linear HMM"""
    hmm = deepcopy(hmm_lr)
    topology = _LinearTopology(n_states=5, random_state=random_state)
    initial = topology.random_initial()
    hmm.initial_ = initial
    assert_equal(hmm.initial_, initial)

def test_ergodic_initial_left_right():
    """Set an initial state distribution generated by an ergodic topology on a left-right HMM"""
    hmm = deepcopy(hmm_e)
    topology = _LeftRightTopology(n_states=5, random_state=random_state)
    initial = topology.random_initial()
    hmm.initial_ = initial
    assert_equal(hmm.initial_, initial)

def test_ergodic_initial_ergodic():
    """Set an initial state distribution generated by an ergodic topology on an ergodic HMM"""
    hmm = deepcopy(hmm_e)
    topology = _ErgodicTopology(n_states=5, random_state=random_state)
    initial = topology.random_initial()
    hmm.initial_ = initial
    assert_equal(hmm.initial_, initial)

def test_ergodic_initial_linear():
    """Set an initial state distribution generated by an ergodic topology on a linear HMM"""
    hmm = deepcopy(hmm_e)
    topology = _LinearTopology(n_states=5, random_state=random_state)
    initial = topology.random_initial()
    hmm.initial_ = initial
    assert_equal(hmm.initial_, initial)

def test_linear_initial_left_right():
    """Set an initial state distribution generated by a linear topology on a left-right HMM"""
    hmm = deepcopy(hmm_lin)
    topology = _LeftRightTopology(n_states=5, random_state=random_state)
    initial = topology.random_initial()
    hmm.initial_ = initial
    assert_equal(hmm.initial_, initial)

def test_linear_initial_ergodic():
    """Set an initial state distribution generated by a linear topology on an ergodic HMM"""
    hmm = deepcopy(hmm_lin)
    topology = _ErgodicTopology(n_states=5, random_state=random_state)
    initial = topology.random_initial()
    hmm.initial_ = initial
    assert_equal(hmm.initial_, initial)

def test_linear_initial_linear():
    """Set an initial state distribution generated by a linear topology on an linear HMM"""
    hmm = deepcopy(hmm_lin)
    topology = _LinearTopology(n_states=5, random_state=random_state)
    initial = topology.random_initial()
    hmm.initial_ = initial
    assert_equal(hmm.initial_, initial)

# ============================ #
# GMMHMM.transitions_ (setter) #
# ============================ #

def test_left_right_transitions_left_right():
    """Set a transition matrix generated by a left-right topology on a left-right HMM"""
    hmm = deepcopy(hmm_lr)
    topology = _LeftRightTopology(n_states=5, random_state=random_state)
    transitions = topology.random_transitions()
    hmm.transitions_ = transitions
    assert_equal(hmm.transitions_, transitions)

def test_left_right_transitions_ergodic():
    """Set a transition matrix generated by a left-right topology on an ergodic HMM"""
    hmm = deepcopy(hmm_lr)
    topology = _ErgodicTopology(n_states=5, random_state=random_state)
    transitions = topology.random_transitions()
    with pytest.raises(ValueError) as e:
        hmm.transitions_ = transitions
    assert str(e.value) == 'Left-right transition matrix must be upper-triangular'

def test_left_right_transitions_linear():
    """Set a transition matrix generated by a left-right topology on a linear HMM"""
    hmm = deepcopy(hmm_lr)
    topology = _LinearTopology(n_states=5, random_state=random_state)
    transitions = topology.random_transitions()
    hmm.transitions_ = transitions
    assert_equal(hmm.transitions_, transitions)

def test_ergodic_transitions_left_right():
    """Set a transition matrix generated by an ergodic topology on a left-right HMM"""
    hmm = deepcopy(hmm_e)
    topology = _LeftRightTopology(n_states=5, random_state=random_state)
    transitions = topology.random_transitions()
    with pytest.warns(UserWarning) as w:
        hmm.transitions_ = transitions
    assert w[0].message.args[0] == 'Zero probabilities in ergodic transition matrix - these transition probabilities will not be learned'
    assert_equal(hmm.transitions_, transitions)

def test_ergodic_transitions_ergodic():
    """Set a transition matrix generated by an ergodic topology on an ergodic HMM"""
    hmm = deepcopy(hmm_e)
    topology = _ErgodicTopology(n_states=5, random_state=random_state)
    transitions = topology.random_transitions()
    hmm.transitions_ = transitions
    assert_equal(hmm.transitions_, transitions)

def test_ergodic_transitions_linear():
    """Set a transition matrix generated by an ergodic topology on a linear HMM"""
    hmm = deepcopy(hmm_e)
    topology = _LinearTopology(n_states=5, random_state=random_state)
    transitions = topology.random_transitions()
    with pytest.warns(UserWarning) as w:
        hmm.transitions_ = transitions
    assert w[0].message.args[0] == 'Zero probabilities in ergodic transition matrix - these transition probabilities will not be learned'
    assert_equal(hmm.transitions_, transitions)

def test_linear_transitions_left_right():
    """Set a transition matrix generated by a linear topology on a left-right HMM"""
    hmm = deepcopy(hmm_lin)
    topology = _LeftRightTopology(n_states=5, random_state=random_state)
    transitions = topology.random_transitions()
    with pytest.raises(ValueError) as e:
        hmm.transitions_ = transitions
    assert str(e.value) == 'Linear transition matrix must only consist of a diagonal and upper diagonal'

def test_linear_transitions_ergodic():
    """Set a transition matrix generated by a linear topology on an ergodic HMM"""
    hmm = deepcopy(hmm_lin)
    topology = _ErgodicTopology(n_states=5, random_state=random_state)
    transitions = topology.random_transitions()
    with pytest.raises(ValueError) as e:
        hmm.transitions_ = transitions
    assert str(e.value) == 'Left-right transition matrix must be upper-triangular'

def test_linear_transitions_linear():
    """Set a transition matrix generated by a linear topology on a linear HMM"""
    hmm = deepcopy(hmm_lin)
    topology = _LinearTopology(n_states=5, random_state=random_state)
    transitions = topology.random_transitions()
    hmm.transitions_ = transitions
    assert_equal(hmm.transitions_, transitions)


import pytest, warnings, os, numpy as np, pickle
from copy import deepcopy
from multiprocessing import cpu_count
from sequentia.classifiers import KNNClassifier
from ....support import assert_equal, assert_all_equal, assert_not_equal

# Set seed for reproducible randomness
seed = 0
np.random.seed(seed)
rng = np.random.RandomState(seed)

classes = ['c0', 'c1', 'c2', 'c3', 'c4']

# Create some sample data
X = [rng.random((10 * i, 3)) for i in range(1, 7)]
y = ['c1', 'c1', 'c0', 'c1', 'c1', 'c0']
x = X[0]

clfs = {
    'k=1': KNNClassifier(k=1, classes=classes, random_state=rng),
    'k=2': KNNClassifier(k=2, classes=classes, random_state=rng),
    'k=3': KNNClassifier(k=3, classes=classes, random_state=rng),
    'weighted': KNNClassifier(k=3, classes=classes, weighting=(lambda x: np.exp(-x)), random_state=rng),
    'independent': KNNClassifier(k=1, classes=classes, independent=True, random_state=rng)
}

for _, clf in clfs.items():
    clf.fit(X, y)

# =================== #
# KNNClassifier.fit() #
# =================== #

def test_fit_sets_attributes():
    """Check that fitting sets the hidden attributes"""
    clf = clfs['k=1']
    clf.fit(X, y)
    assert_all_equal(clf.X_, X)
    assert_equal(clf.y_, clf.encoder_.transform(y))
    assert clf._n_features_ == 3

# ============================= #
# KNNClassifier._multi_argmax() #
# ============================= #

def test_multi_argmax_single():
    """Check that the correct index is returned when there is a single maximum"""
    clf = clfs['k=1']
    idx = clf._multi_argmax([-2, 1, 5, -1, 3, 4, 2])
    assert list(idx) == [2]

def test_multi_argmax_multiple():
    """Check that the correct indicies are returned when there are multiple maxima"""
    clf = clfs['k=1']
    idx = clf._multi_argmax([-2, 1, 5, -1, 3, 5, 4, 2])
    assert list(idx) == [2, 5]

# =============================== #
# KNNClassifier._find_k_nearest() #
# =============================== #

def test_find_k_nearest_tie_within():
    """Check that tied labels are fetched correctly when they are in the nearest k values."""
    clf = deepcopy(clfs['k=3'])
    clf._y_ = clf._encoder_.transform(classes)
    labels, scores = clf._find_k_nearest(np.array([4, 2, 1, 1, 3]))
    assert len(labels) == clf._k
    assert all(label in (1, 2, 3) for label in labels)
    assert all(score == 1 for score in scores)

def test_find_k_nearest_tie_partially_within():
    """Check that a correct subset of the labels are fetched when there are more equidistant values than k."""
    clf = deepcopy(clfs['k=3'])
    clf._y_ = clf._encoder_.transform(classes)
    labels, scores = clf._find_k_nearest(np.array([0, 1, 1, 1, 3]))
    assert len(labels) == clf._k
    assert all(label in (0, 1, 2, 3) for label in labels)
    assert all(score == 1 for score in scores)

def test_find_k_nearest_weighting():
    """Check that the correct scores are returned for the k nearest values when a custom weighting is used."""
    clf = deepcopy(clfs['k=3'])
    clf._weighting = lambda x: np.exp(-x)
    clf._y_ = clf._encoder_.transform(classes)
    labels, scores = clf._find_k_nearest(np.array([4, 2, 1, 5, 3]))
    assert len(labels) == clf._k
    assert all(label in (1, 2, 4) for label in labels)
    assert all(score in (np.exp(-1), np.exp(-2), np.exp(-3)) for score in scores)

# ================================ #
# KNNClassifier._find_max_labels() #
# ================================ #

def test_find_max_labels_uniform_two_classes():
    """Check that the correct single maximum value is returned when a uniform weighting is used with two distinct classes."""
    clf = clfs['k=1']
    max_labels, scores = clf._find_max_labels(
        nearest_labels=np.array([1, 0, 0, 0, 1]),
        nearest_scores=np.array([1, 1, 1, 1, 1])
    )
    assert_equal(max_labels, np.array([0]))
    assert_equal(scores, np.array([3., 2., -np.inf, -np.inf, -np.inf]))

def test_find_max_labels_uniform_two_classes_tie():
    """Check that the correct maximum values are returned when a uniform weighting is used with two distinct tied classes."""
    clf = clfs['k=1']
    max_labels, scores = clf._find_max_labels(
        nearest_labels=np.array([1, 1, 0, 0]),
        nearest_scores=np.array([1, 1, 1, 1])
    )
    assert_equal(max_labels, np.array([0, 1]))
    assert_equal(scores, np.array([2., 2., -np.inf, -np.inf, -np.inf]))

def test_find_max_labels_uniform_many_classes():
    """Check that the correct maximum values are returned when a uniform weighting is used with multiple distinct classes."""
    clf = clfs['k=1']
    max_labels, scores = clf._find_max_labels(
        nearest_labels=np.array([1, 0, 0, 0, 1, 2, 2, 3]),
        nearest_scores=np.array([1, 1, 1, 1, 1, 1, 1, 1])
    )
    assert_equal(max_labels, np.array([0]))
    assert_equal(scores, np.array([3, 2, 2, 1, -np.inf]))

def test_find_max_labels_uniform_many_classes_tie():
    """Check that the correct maximum values are returned when a uniform weighting is used with multiple tied distinct classes."""
    clf = clfs['k=1']
    max_labels, scores = clf._find_max_labels(
        nearest_labels=np.array([1, 0, 0, 1, 2, 2, 3, 3]),
        nearest_scores=np.array([1, 1, 1, 1, 1, 1, 1, 1])
    )
    assert_equal(max_labels, np.array([0, 1, 2, 3]))
    assert_equal(scores, np.array([2, 2, 2, 2, -np.inf]))

def test_find_max_labels_weighted_two_classes():
    """Check that the correct single maximum value is returned when a custom weighting is used with two distinct classes."""
    clf = clfs['k=1']
    max_labels, scores = clf._find_max_labels(
        nearest_labels=np.array([1, 0  , 0, 0, 1]),
        nearest_scores=np.array([1, 0.5, 1, 1, 2])
    )
    assert_equal(max_labels, np.array([1]))
    assert_equal(scores, np.array([2.5, 3, -np.inf, -np.inf, -np.inf]))

def test_find_max_labels_weighted_two_classes_tie():
    """Check that the correct maximum values are returned when a custom weighting is used with two distinct tied classes."""
    clf = clfs['k=1']
    max_labels, scores = clf._find_max_labels(
        nearest_labels=np.array([1, 1  , 0  , 0]),
        nearest_scores=np.array([2, 0.5, 1.5, 1])
    )
    assert_equal(max_labels, np.array([0, 1]))
    assert_equal(scores, np.array([2.5, 2.5, -np.inf, -np.inf, -np.inf]))

def test_find_max_labels_weighted_many_classes():
    """Check that the correct maximum values are returned when a custom weighting is used with multiple distinct classes."""
    clf = clfs['k=1']
    max_labels, scores = clf._find_max_labels(
        nearest_labels=np.array([1, 0  , 0  , 0, 1  , 2  , 2, 3]),
        nearest_scores=np.array([2, 0.5, 1.5, 1, 2.5, 2.5, 1, 3])
    )
    assert_equal(max_labels, np.array([1]))
    assert_equal(scores, np.array([3, 4.5, 3.5, 3, -np.inf]))

def test_find_max_labels_weighted_many_classes_tie():
    """Check that the correct maximum values are returned when a custom weighting is used with multiple tied distinct classes."""
    clf = clfs['k=1']
    max_labels, scores = clf._find_max_labels(
        nearest_labels=np.array([1, 0, 0  , 1  , 2   , 2   , 3  , 3]),
        nearest_scores=np.array([1, 2, 1.5, 2.5, 1.75, 1.75, 0.5, 3])
    )
    assert_equal(max_labels, np.array([0, 1, 2, 3]))
    assert_equal(scores, np.array([3.5, 3.5, 3.5, 3.5, -np.inf]))

# ======================= #
# KNNClassifier.predict() #
# ======================= #

def test_predict_without_fit():
    """Predict without fitting the model"""
    with pytest.raises(RuntimeError) as e:
        KNNClassifier(k=1, classes=classes).predict(x, verbose=False)
    assert str(e.value) == 'The classifier needs to be fitted first'

def test_predict_single_k1():
    """Predict a single observation sequence (k=1)"""
    assert clfs['k=1'].predict(x, verbose=False) == 'c1'

def test_predict_single_k2():
    """Predict a single observation sequence (k=2)"""
    assert clfs['k=2'].predict(x, verbose=False) == 'c1'

def test_predict_single_k3():
    """Predict a single observation sequence (k=3)"""
    assert clfs['k=3'].predict(x, verbose=False) == 'c1'

def test_predict_single_weighted():
    """Predict a single observation sequence (weighted)"""
    assert clfs['weighted'].predict(x, verbose=False) == 'c1'

def test_predict_single_independent():
    """Predict a single observation sequence with independent warping"""
    assert clfs['independent'].predict(x, verbose=False) == 'c1'

def test_predict_multiple_k1():
    """Predict multiple observation sequences (k=1)"""
    predictions = clfs['k=1'].predict(X, verbose=False)
    assert list(predictions) == ['c1', 'c1', 'c0', 'c1', 'c1', 'c0']

def test_predict_multiple_k2():
    """Predict multiple observation sequences (k=2)"""
    predictions = clfs['k=2'].predict(X, verbose=False)
    assert list(predictions) == ['c1', 'c1', 'c1', 'c1', 'c1', 'c1']

def test_predict_multiple_k3():
    """Predict multiple observation sequences (k=3)"""
    predictions = clfs['k=3'].predict(X, verbose=False)
    assert list(predictions) == ['c1', 'c1', 'c1', 'c1', 'c1', 'c0']

def test_predict_multiple_weighted():
    """Predict multiple observation sequences (weighted)"""
    predictions = clfs['weighted'].predict(X, verbose=False)
    assert list(predictions) == ['c1', 'c1', 'c0', 'c1', 'c1', 'c0']

def test_predict_multiple_independent():
    """Predict multiple observation sequences with independent warping"""
    predictions = clfs['independent'].predict(X, verbose=False)
    assert list(predictions) == ['c1', 'c1', 'c0', 'c1', 'c1', 'c0']

def test_predict_single_encoded_label():
    """Predict a single observation sequence and return the encoded labels"""
    assert clfs['k=3'].predict(x, original_labels=False, verbose=False) == 1

def test_predict_single_original_label():
    """Predict a single observation sequence and return the original labels"""
    assert clfs['k=3'].predict(x, original_labels=True, verbose=False) == 'c1'

def test_predict_multiple_encoded_labels():
    """Predict multiple observation sequences and return the encoded labels"""
    predictions = clfs['k=3'].predict(X, original_labels=False, verbose=False)
    assert list(predictions) == [1, 1, 1, 1, 1, 0]

def test_predict_multiple_original_labels():
    """Predict multiple observation sequences and return the original labels"""
    predictions = clfs['k=3'].predict(X, original_labels=True, verbose=False)
    assert list(predictions) == ['c1', 'c1', 'c1', 'c1', 'c1', 'c0']

def test_predict_single_k1_with_scores():
    """Predict a single observation sequence (k=1) with scores"""
    label, scores = clfs['k=1'].predict(x, return_scores=True, verbose=False)
    assert label == 'c1'
    assert_equal(scores, np.array([-np.inf, 1., -np.inf, -np.inf, -np.inf]))

def test_predict_single_k2_with_scores():
    """Predict a single observation sequence (k=2) with scores"""
    label, scores = clfs['k=2'].predict(x, return_scores=True, verbose=False)
    assert label == 'c1'
    assert_equal(scores, np.array([-np.inf, 2., -np.inf, -np.inf, -np.inf]))

def test_predict_single_k3_with_scores():
    """Predict a single observation sequence (k=3) with scores"""
    label, scores = clfs['k=3'].predict(x, return_scores=True, verbose=False)
    assert label == 'c1'
    assert_equal(scores, np.array([1., 2., -np.inf, -np.inf, -np.inf]))

def test_predict_single_weighted_with_scores():
    """Predict a single observation sequence (weighted) with scores"""
    label, scores = clfs['weighted'].predict(x, return_scores=True, verbose=False)
    assert label == 'c1'
    assert_equal(scores, np.array([0.0719717, 1.08412051, -np.inf, -np.inf, -np.inf]))

def test_predict_single_independent_with_scores():
    """Predict a single observation sequence with independent warping with scores"""
    label, scores = clfs['independent'].predict(x, return_scores=True, verbose=False)
    assert label == 'c1'
    assert_equal(scores, np.array([-np.inf, 1., -np.inf, -np.inf, -np.inf]))

def test_predict_multiple_k1_with_scores():
    """Predict multiple observation sequences (k=1) with scores"""
    labels, scores = clfs['k=1'].predict(X, return_scores=True, verbose=False)
    assert list(labels) == ['c1', 'c1', 'c0', 'c1', 'c1', 'c0']
    assert_equal(scores, np.array([
        [-np.inf,      1., -np.inf, -np.inf, -np.inf],
        [-np.inf,      1., -np.inf, -np.inf, -np.inf],
        [     1., -np.inf, -np.inf, -np.inf, -np.inf],
        [-np.inf,      1., -np.inf, -np.inf, -np.inf],
        [-np.inf,      1., -np.inf, -np.inf, -np.inf],
        [     1., -np.inf, -np.inf, -np.inf, -np.inf]
    ]))

def test_predict_multiple_k2_with_scores():
    """Predict multiple observation sequences (k=2) with scores"""
    labels, scores = clfs['k=2'].predict(X, return_scores=True, verbose=False)
    assert list(labels) == ['c1', 'c1', 'c1', 'c1', 'c0', 'c0']
    assert_equal(scores, np.array([
        [-np.inf, 2., -np.inf, -np.inf, -np.inf],
        [-np.inf, 2., -np.inf, -np.inf, -np.inf],
        [     1., 1., -np.inf, -np.inf, -np.inf],
        [-np.inf, 2., -np.inf, -np.inf, -np.inf],
        [     1., 1., -np.inf, -np.inf, -np.inf],
        [     1., 1., -np.inf, -np.inf, -np.inf]
    ]))

def test_predict_multiple_k3_with_scores():
    """Predict multiple observation sequences (k=3) with scores"""
    labels, scores = clfs['k=3'].predict(X, return_scores=True, verbose=False)
    assert list(labels) == ['c1', 'c1', 'c1', 'c1', 'c1', 'c0']
    assert_equal(scores, np.array([
        [     1., 2., -np.inf, -np.inf, -np.inf],
        [     1., 2., -np.inf, -np.inf, -np.inf],
        [     1., 2., -np.inf, -np.inf, -np.inf],
        [-np.inf, 3., -np.inf, -np.inf, -np.inf],
        [     1., 2., -np.inf, -np.inf, -np.inf],
        [     2., 1., -np.inf, -np.inf, -np.inf]
    ]))

def test_predict_multiple_weighted_with_scores():
    """Predict multiple observation sequences (weighted) with scores"""
    labels, scores = clfs['weighted'].predict(X, return_scores=True, verbose=False)
    assert list(labels) == ['c1', 'c1', 'c0', 'c1', 'c1', 'c0']
    assert_equal(scores, np.array([
        [0.0719717 , 1.08412051, -np.inf, -np.inf, -np.inf],
        [0.04549809, 1.08412051, -np.inf, -np.inf, -np.inf],
        [        1., 0.1174698 , -np.inf, -np.inf, -np.inf],
        [   -np.inf, 1.07904269, -np.inf, -np.inf, -np.inf],
        [0.02438639, 1.02240087, -np.inf, -np.inf, -np.inf],
        [1.01869848, 0.01882683, -np.inf, -np.inf, -np.inf]
    ]))

def test_predict_multiple_independent_with_scores():
    """Predict multiple observation sequences with independent warping with scores"""
    labels, scores = clfs['independent'].predict(X, return_scores=True, verbose=False)
    assert list(labels) == ['c1', 'c1', 'c0', 'c1', 'c1', 'c0']
    assert_equal(scores, np.array([
        [-np.inf,      1., -np.inf, -np.inf, -np.inf],
        [-np.inf,      1., -np.inf, -np.inf, -np.inf],
        [     1., -np.inf, -np.inf, -np.inf, -np.inf],
        [-np.inf,      1., -np.inf, -np.inf, -np.inf],
        [-np.inf,      1., -np.inf, -np.inf, -np.inf],
        [     1., -np.inf, -np.inf, -np.inf, -np.inf]
    ]))

def test_predict_single_encoded_label_with_scores():
    """Predict a single observation sequence and return the encoded labels with scores"""
    label, scores = clfs['k=3'].predict(x, return_scores=True, original_labels=False, verbose=False)
    assert label == 1
    assert_equal(scores, np.array([1., 2., -np.inf, -np.inf, -np.inf]))

def test_predict_single_original_label_with_scores():
    """Predict a single observation sequence and return the original labels with scores"""
    label, scores = clfs['k=3'].predict(x, return_scores=True, original_labels=True, verbose=False)
    assert label == 'c1'
    assert_equal(scores, np.array([1., 2., -np.inf, -np.inf, -np.inf]))

def test_predict_multiple_encoded_labels_with_scores():
    """Predict multiple observation sequences and return the encoded labels"""
    labels, scores = clfs['k=3'].predict(X, return_scores=True, original_labels=False, verbose=False)
    assert list(labels) == [1, 1, 1, 1, 1, 0]
    assert_equal(scores, np.array([
        [     1., 2., -np.inf, -np.inf, -np.inf],
        [     1., 2., -np.inf, -np.inf, -np.inf],
        [     1., 2., -np.inf, -np.inf, -np.inf],
        [-np.inf, 3., -np.inf, -np.inf, -np.inf],
        [     1., 2., -np.inf, -np.inf, -np.inf],
        [     2., 1., -np.inf, -np.inf, -np.inf]
    ]))

def test_predict_multiple_original_labels_with_scores():
    """Predict multiple observation sequences and return the original labels with scores"""
    labels, scores = clfs['k=3'].predict(X, return_scores=True, original_labels=True, verbose=False)
    assert list(labels) == ['c1', 'c1', 'c1', 'c1', 'c1', 'c0']
    assert_equal(scores, np.array([
        [     1., 2., -np.inf, -np.inf, -np.inf],
        [     1., 2., -np.inf, -np.inf, -np.inf],
        [     1., 2., -np.inf, -np.inf, -np.inf],
        [-np.inf, 3., -np.inf, -np.inf, -np.inf],
        [     1., 2., -np.inf, -np.inf, -np.inf],
        [     2., 1., -np.inf, -np.inf, -np.inf]
    ]))

# ======================== #
# KNNClassifier.evaluate() #
# ======================== #

def test_evaluate():
    """Evaluate performance on some observation sequences and labels"""
    acc, cm = clfs['k=3'].evaluate(X, y)
    assert acc == 5 / 6
    assert_equal(cm, np.array([
        [1, 1, 0, 0, 0],
        [0, 4, 0, 0, 0],
        [0, 0, 0, 0, 0],
        [0, 0, 0, 0, 0],
        [0, 0, 0, 0, 0]
    ]))

# ==================== #
# KNNClassifier.save() #
# ==================== #

def test_save_unfitted():
    """Save an unfitted KNNClassifier object."""
    try:
        with pytest.raises(RuntimeError) as e:
            KNNClassifier(k=1, classes=classes).save('test.pkl')
        assert str(e.value) == 'The classifier needs to be fitted first'
    finally:
        if os.path.exists('test.pkl'):
            os.remove('test.pkl')

def test_save_fitted():
    """Save a fitted KNNClassifier object."""
    try:
        clfs['weighted'].save('test.pkl')
        assert os.path.isfile('test.pkl')
    finally:
        os.remove('test.pkl')

# ==================== #
# KNNClassifier.load() #
# ==================== #

def test_load_invalid_format():
    """Load a KNNClassifier from an illegally formatted file"""
    try:
        with open('test', 'w') as f:
            f.write('illegal')
        with pytest.raises(pickle.UnpicklingError) as e:
            KNNClassifier.load('test')
    finally:
        os.remove('test')

def test_load_invalid_object_type():
    """Load a KNNClassifier from a pickle file with invalid object type"""
    try:
        with open('test.pkl', 'wb') as file:
            pickle.dump(0, file)
        with pytest.raises(TypeError) as e:
            KNNClassifier.load('test.pkl')
        assert str(e.value) == 'Expected deserialized object to be a dictionary - make sure the object was serialized with the save() function'
    finally:
        os.remove('test.pkl')

def test_load_missing_keys():
    """Load a KNNClassifier from a pickled dict with invalid keys"""
    try:
        with open('test.pkl', 'wb') as file:
            pickle.dump({'test': 0}, file)
        with pytest.raises(ValueError) as e:
            KNNClassifier.load('test.pkl')
        assert str(e.value) == 'Missing keys in deserialized object dictionary – make sure the object was serialized with the save() function'
    finally:
        os.remove('test.pkl')

def test_load_valid_no_weighting():
    """Load a serialized KNNClassifier with the default weighting function"""
    try:
        clfs['k=3'].save('test.pkl')
        clf = KNNClassifier.load('test.pkl')
        # Check that all fields are still the same
        assert isinstance(clf, KNNClassifier)
        assert clf._k == 3
        assert list(clf.encoder_.classes_) == classes
        assert clf._window == 1.
        assert clf._use_c == False
        assert clf._independent == False
        assert deepcopy(clf._random_state).normal() == deepcopy(rng).normal()
        assert_all_equal(clf.X_, X)
        assert_equal(clf.y_, clf.encoder_.transform(y))
        assert clf._n_features_ == 3
        # Check that weighting functions are the same for x=0 to x=1000
        xs = np.arange(1000, step=0.1)
        weighting = lambda x: np.ones(x.size)
        assert_equal(clf._weighting(xs), weighting(xs))
    finally:
        os.remove('test.pkl')

def test_load_valid_weighting():
    """Load a serialized KNNClassifier with a custom weighting function"""
    try:
        clfs['weighted'].save('test.pkl')
        clf = KNNClassifier.load('test.pkl')
        # Check that all fields are still the same
        assert isinstance(clf, KNNClassifier)
        assert clf._k == 3
        assert list(clf.encoder_.classes_) == classes
        assert clf._window == 1.
        assert clf._use_c == False
        assert clf._independent == False
        assert deepcopy(clf._random_state).normal() == deepcopy(rng).normal()
        assert_all_equal(clf.X_, X)
        assert_equal(clf.y_, clf.encoder_.transform(y))
        assert clf._n_features_ == 3
        # Check that weighting functions are the same for x=0 to x=1000
        xs = np.arange(1000, step=0.1)
        weighting = lambda x: np.exp(-x)
        assert_equal(clf._weighting(xs), weighting(xs))
    finally:
        os.remove('test.pkl')


import numpy as np
from pkg_resources import resource_filename
from .base import Dataset
from ..internals import _Validator

def load_digits(numbers=range(10), random_state=None):
    """Load audio samples of spoken digits from the Free Spoken Digit Dataset.

    The `Free Spoken Digit Dataset (FSDD) <https://github.com/Jakobovski/free-spoken-digit-dataset>`_
    consists of 3000 recordings of the spoken digits 0-9.

    The dataset consists of 50 recordings of each digit by 6 individual speakers.

    Parameters
    ----------
    numbers: array-like of int
        Subset of digits to include in the dataset. Defaults to 0-9.

    random_state: numpy.random.RandomState, int, optional
        A random state object or seed for reproducible randomness.

    Returns
    -------
    dataset: :class:`sequentia.datasets.Dataset`
        A dataset object representing the loaded digits.
    """
    random_state = _Validator().is_random_state(random_state)

    # Load the dataset from compressed numpy file
    data = np.load(resource_filename('sequentia', 'datasets/data/digits.npz'))

    # Fetch arrays from loaded file
    sequences, lengths, labels = data['sequences'], data['lengths'], data['labels']

    # Find the starts and ends of each sequence
    starts = np.zeros_like(lengths)
    starts[1:] = np.cumsum(lengths[:-1])
    ends = np.cumsum(lengths)

    # Split the array into sequences
    X = [sequences[start:end] for start, end in zip(starts, ends)]

    # Select and create a Dataset only with sequences having the specified labels
    X_np = np.array(X, dtype=object)
    idx = np.argwhere(np.isin(labels, numbers)).flatten()
    return Dataset(X_np[idx].tolist(), labels[idx], numbers, random_state)